﻿/*
 * nssudeaudio - USB Bus Class Filter Driver for fixing UDE compatibility with USBAUDIO.SYS
 *
 * Copyright (c) 2022-2023, Nefarius Software Solutions e.U.
 *  All rights reserved.
 *
 *  Redistribution and use in bytecode or binary forms, without modification,
 *  are permitted provided that the following conditions are met:
 *
 *    * Redistributions in bytecode or binary form must reproduce the above
 *      copyright notice, this list of conditions and the following disclaimer
 *      in the documentation and/or other materials provided with the
 *      distribution. This includes repository READMEs for open-source projects
 *      interfacing with this product, websites or in-application credits.
 *
 *    * Neither the name of the nssudeaudio.sys nor the names of its contributors may
 *      be used to endorse or promote products derived from or distributed
 *      with this software without specific prior written permission.
 *
 *    * Bundling or otherwise redistributing copies of binaries signed by
 *      Nefarius Software Solutions e.U. is prohibited without specific prior
 *      written permission.
 *
 *    * Personal use for research, experiment, and testing for
 *      the benefit of public knowledge, personal study, private
 *      entertainment, hobby projects, amateur pursuits, or religious
 *      observance, without any anticipated commercial application,
 *      is use for a permitted purpose.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */


#include "Driver.h"
#include "Device.tmh"

#include <usbbusif.h>


static
VOID
UsbInterfaceReference(
	_In_ PVOID Context
)
{
	const PCHILD_DEVICE_CONTEXT pCtx = ChildDeviceGetContext(Context);

	pCtx->OriginalInterface.InterfaceReference(pCtx->OriginalInterface.BusContext);
}

static
VOID
UsbInterfaceDereference(
	_In_ PVOID Context
)
{
	const PCHILD_DEVICE_CONTEXT pCtx = ChildDeviceGetContext(Context);

	pCtx->OriginalInterface.InterfaceDereference(pCtx->OriginalInterface.BusContext);
}

static
NTSTATUS
USB_BUSIFFN
UsbInterfaceQueryBusTime(
	IN PVOID BusContext,
	IN OUT PULONG CurrentUsbFrame
)
{
	FuncEntry(TRACE_DEVICE);

	const PCHILD_DEVICE_CONTEXT pCtx = ChildDeviceGetContext(BusContext);

	NTSTATUS status = pCtx->OriginalInterface.QueryBusTime(
		pCtx->OriginalInterface.BusContext,
		CurrentUsbFrame
	);

	if (CurrentUsbFrame)
	{
		TraceInformation(
			TRACE_DEVICE,
			"UsbInterfaceQueryBusTime.CurrentUsbFrame = %d",
			*CurrentUsbFrame
		);
	}

	// fake-succeed the call
	if (CurrentUsbFrame && status == STATUS_NOT_SUPPORTED)
	{
		*CurrentUsbFrame = 0;
		status = STATUS_SUCCESS;
	}

	FuncExit(TRACE_DEVICE, "status=%!STATUS!", status);

	return status;
}

static
NTSTATUS
USB_BUSIFFN
UsbInterfaceSubmitIsoOutUrb(
	IN PVOID BusContext,
	IN PURB Urb
)
{
	const PCHILD_DEVICE_CONTEXT pCtx = ChildDeviceGetContext(BusContext);

	return pCtx->OriginalInterface.SubmitIsoOutUrb(
		pCtx->OriginalInterface.BusContext,
		Urb
	);
}

BOOLEAN
USB_BUSIFFN
UsbInterfaceIsDeviceHighSpeed(
	IN PVOID BusContext
)
{
	const PCHILD_DEVICE_CONTEXT pCtx = ChildDeviceGetContext(BusContext);

	return pCtx->OriginalInterface.IsDeviceHighSpeed(
		pCtx->OriginalInterface.BusContext
	);
}

NTSTATUS
USB_BUSIFFN
UsbInterfaceQueryBusInformation(
	IN PVOID BusContext,
	IN ULONG Level,
	IN OUT PVOID BusInformationBuffer,
	IN OUT PULONG BusInformationBufferLength,
	OUT PULONG BusInformationActualLength
)
{
	const PCHILD_DEVICE_CONTEXT pCtx = ChildDeviceGetContext(BusContext);

	return pCtx->OriginalInterface.QueryBusInformation(
		pCtx->OriginalInterface.BusContext,
		Level,
		BusInformationBuffer,
		BusInformationBufferLength,
		BusInformationActualLength
	);
}

VOID
USB_BUSIFFN
UsbInterfaceGetUSBDIVersion(
	IN PVOID BusContext,
	IN OUT PUSBD_VERSION_INFORMATION VersionInformation,
	IN OUT PULONG HcdCapabilities
)
{
	const PCHILD_DEVICE_CONTEXT pCtx = ChildDeviceGetContext(BusContext);

	pCtx->OriginalInterface.GetUSBDIVersion(
		pCtx->OriginalInterface.BusContext,
		VersionInformation,
		HcdCapabilities
	);
}

NTSTATUS
EvtChildDeviceAdded(
	_In_ WDFDEVICE Device,
	_In_ DMFBUSCHILDDEVICE ChildDevice
)
{
	NTSTATUS status;
	WDF_OBJECT_ATTRIBUTES attributes;
	PCHILD_DEVICE_CONTEXT childContext = NULL;

	FuncEntry(TRACE_DEVICE);

	WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attributes, CHILD_DEVICE_CONTEXT);

	if (NT_SUCCESS(status = WdfObjectAllocateContext(
		ChildDevice,
		&attributes,
		(void**)&childContext
	)))
	{
		childContext->Parent = Device;
	}
	else
	{
		TraceError(
			TRACE_DEVICE,
			"WdfObjectAllocateContext failed with status %!STATUS!",
			status
		);
		EventWriteFailedWithNTStatus(NULL, __FUNCTION__, L"WdfObjectAllocateContext", status);
	}

	FuncExit(TRACE_DEVICE, "status=%!STATUS!", status);

	return status;
}

BOOLEAN
EvtChildDeviceQueryInterface(
	_In_ DMFBUSCHILDDEVICE ChildDevice,
	_In_ PIRP Irp
)
{
	FuncEntry(TRACE_DEVICE);

	const PCHILD_DEVICE_CONTEXT pCtx = ChildDeviceGetContext(ChildDevice);
	const PIO_STACK_LOCATION currentStack = IoGetCurrentIrpStackLocation(Irp);

	if (KeGetCurrentIrql() > PASSIVE_LEVEL)
	{
		return FALSE;
	}

	if (!pCtx)
	{
		TraceError(
			TRACE_DEVICE,
			"Child device context not yet allocated"
		);
		return FALSE;
	}

	// skip not interested
	if (!IsEqualGUID(currentStack->Parameters.QueryInterface.InterfaceType, &USB_BUS_INTERFACE_USBDI_GUID) ||
		currentStack->Parameters.QueryInterface.Version != 1)
	{
		return FALSE;
	}

	PUSB_BUS_INTERFACE_USBDI_V1 pUpperInterface = (PUSB_BUS_INTERFACE_USBDI_V1)currentStack->Parameters.QueryInterface.
		Interface;

	TraceInformation(
		TRACE_DEVICE,
		"PRE QueryInterface.InterfaceType = %!GUID!",
		currentStack->Parameters.QueryInterface.InterfaceType
	);

	TraceInformation(
		TRACE_DEVICE,
		"PRE QueryInterface.Version = %d",
		currentStack->Parameters.QueryInterface.Version
	);

	KEVENT ke;
	IO_STATUS_BLOCK iosb;

	RtlZeroMemory(&iosb, sizeof(IO_STATUS_BLOCK));

	PUSB_BUS_INTERFACE_USBDI_V1 pUsbInterface = ExAllocatePool2(
		PagedPool,
		sizeof(USB_BUS_INTERFACE_USBDI_V1),
		NSS_POOL_TAG
	);

	if (!pUsbInterface)
	{
		return FALSE;
	}

	RtlZeroMemory(pUsbInterface, sizeof(USB_BUS_INTERFACE_USBDI_V1));

	const PDEVICE_OBJECT deviceObject = DMF_BusFilter_WdmAttachedDeviceGet(ChildDevice);

	KeInitializeEvent(&ke, NotificationEvent, FALSE);

	const PIRP qiIrp = IoBuildSynchronousFsdRequest(
		IRP_MJ_PNP,
		deviceObject,
		NULL,
		0,
		NULL,
		&ke,
		&iosb
	);
	qiIrp->IoStatus.Status = STATUS_NOT_SUPPORTED; // required initialize
	qiIrp->IoStatus.Information = 0;

	const PIO_STACK_LOCATION qiStack = IoGetNextIrpStackLocation(qiIrp);

	qiStack->MajorFunction = IRP_MJ_PNP;
	qiStack->MinorFunction = IRP_MN_QUERY_INTERFACE;

	qiStack->Parameters.QueryInterface.InterfaceType = currentStack->Parameters.QueryInterface.InterfaceType;
	qiStack->Parameters.QueryInterface.Version = currentStack->Parameters.QueryInterface.Version;
	qiStack->Parameters.QueryInterface.Size = currentStack->Parameters.QueryInterface.Size;
	qiStack->Parameters.QueryInterface.Interface = (PINTERFACE)pUsbInterface;
	qiStack->Parameters.QueryInterface.InterfaceSpecificData = NULL;

	const NTSTATUS status = IoCallDriver(deviceObject, qiIrp);

	if (status == STATUS_PENDING)
	{
		KeWaitForSingleObject(&ke, Executive, KernelMode, FALSE, NULL);
	}

	BOOLEAN ret = FALSE;

	if (NT_SUCCESS(qiIrp->IoStatus.Status))
	{
		TraceInformation(
			TRACE_DEVICE,
			"IoCallDriver succeeded with status %!STATUS!",
			qiIrp->IoStatus.Status
		);

		TraceInformation(
			TRACE_DEVICE,
			"POST usbInterface.Version = %d",
			pUsbInterface->Version
		);

		TraceInformation(
			TRACE_DEVICE,
			"POST usbInterface.Size = %d",
			pUsbInterface->Size
		);

		pCtx->OriginalInterface.Size = pUpperInterface->Size = pUsbInterface->Size;
		pCtx->OriginalInterface.Version = pUpperInterface->Version = pUsbInterface->Version;

		// save original functions
		pCtx->OriginalInterface.InterfaceReference = pUsbInterface->InterfaceReference;
		pCtx->OriginalInterface.InterfaceDereference = pUsbInterface->InterfaceDereference;
		pCtx->OriginalInterface.GetUSBDIVersion = pUsbInterface->GetUSBDIVersion;
		pCtx->OriginalInterface.SubmitIsoOutUrb = pUsbInterface->SubmitIsoOutUrb;
		pCtx->OriginalInterface.QueryBusInformation = pUsbInterface->QueryBusInformation;
		pCtx->OriginalInterface.IsDeviceHighSpeed = pUsbInterface->IsDeviceHighSpeed;
		pCtx->OriginalInterface.QueryBusTime = pUsbInterface->QueryBusTime;

		// report proxy functions
		pUpperInterface->InterfaceReference = UsbInterfaceReference;
		pUpperInterface->InterfaceDereference = UsbInterfaceDereference;
		pUpperInterface->GetUSBDIVersion = UsbInterfaceGetUSBDIVersion;
		pUpperInterface->SubmitIsoOutUrb = UsbInterfaceSubmitIsoOutUrb;
		pUpperInterface->QueryBusInformation = UsbInterfaceQueryBusInformation;
		pUpperInterface->IsDeviceHighSpeed = UsbInterfaceIsDeviceHighSpeed;
		pUpperInterface->QueryBusTime = UsbInterfaceQueryBusTime;

		// keep original bus context
		pCtx->OriginalInterface.BusContext = pUsbInterface->BusContext;
		// replace context object so we get it in our hooked function
		pUpperInterface->BusContext = ChildDevice;

		Irp->IoStatus.Status = STATUS_SUCCESS;
		// causes the module to complete the request
		ret = TRUE;
	}

	ExFreePoolWithTag(pUsbInterface, NSS_POOL_TAG);

	FuncExitNoReturn(TRACE_DEVICE);
	return ret;
}
