﻿/*
 * nssudeaudio - USB Bus Class Filter Driver for fixing UDE compatibility with USBAUDIO.SYS
 *
 * Copyright (c) 2022-2023, Nefarius Software Solutions e.U.
 *  All rights reserved.
 *
 *  Redistribution and use in bytecode or binary forms, without modification,
 *  are permitted provided that the following conditions are met:
 *
 *    * Redistributions in bytecode or binary form must reproduce the above
 *      copyright notice, this list of conditions and the following disclaimer
 *      in the documentation and/or other materials provided with the
 *      distribution. This includes repository READMEs for open-source projects
 *      interfacing with this product, websites or in-application credits.
 *
 *    * Neither the name of the nssudeaudio.sys nor the names of its contributors may
 *      be used to endorse or promote products derived from or distributed
 *      with this software without specific prior written permission.
 *
 *    * Bundling or otherwise redistributing copies of binaries signed by
 *      Nefarius Software Solutions e.U. is prohibited without specific prior
 *      written permission.
 *
 *    * Personal use for research, experiment, and testing for
 *      the benefit of public knowledge, personal study, private
 *      entertainment, hobby projects, amateur pursuits, or religious
 *      observance, without any anticipated commercial application,
 *      is use for a permitted purpose.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */


#include "Driver.h"
#include "Util.tmh"


 //
 // Set a device property on the provided PDO
 // 
#pragma code_seg("PAGE")
_IRQL_requires_max_(PASSIVE_LEVEL)
NTSTATUS
SetDeviceProperty(
	_In_ DMFBUSCHILDDEVICE Device,
	_In_ const DEVPROPKEY* PropertyKey,
	_In_ DEVPROPTYPE PropertyType,
	_In_ PWSTR Buffer
)
{
	FuncEntry(TRACE_UTIL);

	NTSTATUS status = STATUS_SUCCESS;
	const PDEVICE_OBJECT pdo = DMF_BusFilter_WdmPhysicalDeviceGet(Device);

	PAGED_CODE();

	do {
		if (!pdo)
		{
			TraceError(
				TRACE_UTIL,
				"Failed to get PDO device object"
			);
			status = STATUS_INVALID_DEVICE_REQUEST;
			break;
		}

		size_t numChars = 0;
		WCHAR* retval = (WCHAR*)Buffer;
		WCHAR* szIter = (WCHAR*)Buffer;

		switch (PropertyType)
		{
		case DEVPROP_TYPE_STRING_LIST:

			while (szIter && *szIter && numChars + wcslen(szIter) + 2 <= MAX_ID_SIZE)
			{
				numChars += wcslen(szIter) + 1;
				szIter += wcslen(szIter) + 1;
			}

			numChars++;

			break;
		case DEVPROP_TYPE_STRING:

			numChars = wcslen(retval) + 1;

			break;
		default:
			status = STATUS_NOT_SUPPORTED;
			break;
		}

		if (!NT_SUCCESS(status))
		{
			break;
		}

		status = IoSetDevicePropertyData(
			pdo,
			PropertyKey,
			LOCALE_NEUTRAL,
			0,
			PropertyType,
			(ULONG)numChars * sizeof(WCHAR),
			(PVOID)Buffer
		);
	} while (FALSE);

	FuncExit(TRACE_UTIL, "status=%!STATUS!", status);

	return status;
}
#pragma code_seg()

//
// Helper function to query device identifiers
// 
#pragma code_seg("PAGE")
_IRQL_requires_max_(PASSIVE_LEVEL)
NTSTATUS BusQueryId(
	PDEVICE_OBJECT DeviceObject,
	BUS_QUERY_ID_TYPE IdType,
	PWCHAR Buffer,
	ULONG BufferLength
)
{
	NTSTATUS status;
	KEVENT ke;
	IO_STATUS_BLOCK iosb;
	PIRP irp;
	PIO_STACK_LOCATION stack;

	FuncEntry(TRACE_UTIL);

	PAGED_CODE();

	const ULONG maxNumChars = BufferLength / sizeof(WCHAR);

	KeInitializeEvent(&ke, NotificationEvent, FALSE);

	RtlZeroMemory(&iosb, sizeof(IO_STATUS_BLOCK));
	irp = IoBuildSynchronousFsdRequest(
		IRP_MJ_PNP,
		DeviceObject,
		NULL,
		0,
		NULL,
		&ke,
		&iosb
	);
	irp->IoStatus.Status = STATUS_NOT_SUPPORTED; // required initialize
	stack = IoGetNextIrpStackLocation(irp);
	stack->MinorFunction = IRP_MN_QUERY_ID;
	stack->Parameters.QueryId.IdType = IdType;

	status = IoCallDriver(DeviceObject, irp);

	if (status == STATUS_PENDING)
	{
		// 
		// Normally, we will not hit this, because QueryId should not be an expensive operation
		// 
		KeWaitForSingleObject(&ke, Executive, KernelMode, FALSE, NULL);
	}

	if (NT_SUCCESS(status))
	{
		if (iosb.Information)
		{
			size_t length = 0;
			WCHAR* retval = (WCHAR*)iosb.Information;
			WCHAR* szIter = (WCHAR*)iosb.Information;

			switch (IdType)  // NOLINT(clang-diagnostic-switch-enum)
			{
				//
				// Double-null terminated multi-strings
				// 
			case BusQueryHardwareIDs:
			case BusQueryCompatibleIDs:

				//
				// Calculate size in characters by finding last double-NULL characters
				// 
				while (szIter && *szIter && length + wcslen(szIter) + 2 <= maxNumChars)
				{
					length += wcslen(szIter) + 1;
					szIter += wcslen(szIter) + 1;
				}

				length++;

				RtlCopyMemory(Buffer, retval, length * sizeof(WCHAR));
				ExFreePool(retval); // IRP_MN_QUERY_ID requires this
				break;
				//
				// Single line string
				// 
			default:
				if (wcslen(retval) + 1 > maxNumChars)
				{
					ExFreePool(retval); // IRP_MN_QUERY_ID requires this
					status = STATUS_BUFFER_TOO_SMALL;
					break;
				}

				RtlStringCchCopyNW(Buffer, maxNumChars, retval, wcslen(retval) + 1);
				ExFreePool(retval); // IRP_MN_QUERY_ID requires this
				break;
			}
		}
		else
		{
			status = STATUS_INVALID_DEVICE_REQUEST;
		}
	}

	FuncExit(TRACE_UTIL, "status=%!STATUS!", status);

	return status;
}
#pragma code_seg()

//
// Retrieves the device instance ID of the PDO
// 
#pragma code_seg("PAGE")
_IRQL_requires_max_(PASSIVE_LEVEL)
NTSTATUS
GetDeviceInstanceIdProperty(
	_In_ DMFBUSCHILDDEVICE ChildDevice,
	_Out_ WDFMEMORY* Value
)
{
	NTSTATUS status = STATUS_NOT_IMPLEMENTED;
	WCHAR buffer[MAX_ID_CHARS];
	ULONG requiredSize = 0;
	DEVPROPTYPE type = 0;
	WDF_OBJECT_ATTRIBUTES attributes;
	PUCHAR memBuffer = NULL;

	FuncEntry(TRACE_UTIL);

	PAGED_CODE();

	PDEVICE_OBJECT pdo = DMF_BusFilter_WdmPhysicalDeviceGet(ChildDevice);

	do
	{
		if (!NT_SUCCESS(status = IoGetDevicePropertyData(
			pdo,
			&DEVPKEY_Device_InstanceId,
			LOCALE_NEUTRAL,
			0,
			MAX_ID_CHARS,
			buffer,
			&requiredSize,
			&type
		)))
		{
			TraceError(
				TRACE_UTIL,
				"IoGetDevicePropertyData failed with status %!STATUS!",
				status
			);
			break;
		}

		WDF_OBJECT_ATTRIBUTES_INIT(&attributes);
		attributes.ParentObject = ChildDevice;

		//
		// Create context memory
		// 
		if (!NT_SUCCESS(status = WdfMemoryCreate(
			&attributes,
			NonPagedPoolNx,
			NSS_POOL_TAG,
			requiredSize,
			Value,
			(void**)&memBuffer
		)))
		{
			TraceError(
				TRACE_UTIL,
				"WdfMemoryCreate failed with status %!STATUS!",
				status
			);
			break;
		}

		RtlCopyMemory(memBuffer, buffer, requiredSize);

	} while (FALSE);

	FuncExit(TRACE_UTIL, "status=%!STATUS!", status);

	return status;
}
#pragma code_seg()
